@@ -7,7 +7,7 @@ from email.mime.image import MIMEImage
 from email.mime.multipart import MIMEMultipart
 from email.mime.text import MIMEText
 from email import encoders
-import inspect
+from inspect import signature, getattr_static
 from itertools import chain
 import json
 from multiprocessing import Pool
@@ -19,6 +19,7 @@ import re
 import smtplib
 import sys
 from tqdm.auto import tqdm
+import wordninja as wn
 
 from htools.config import get_credentials, get_default_user
 
@@ -117,7 +118,7 @@ def hasarg(func, arg):
     >>> hasarg(foo, 'c')
     False
     """
-    return arg in inspect.signature(func).parameters
+    return arg in signature(func).parameters
 
 
 def quickmail(subject, message, to_email, from_email=None, img_path=None,
@@ -328,7 +329,7 @@ def _read_write_args(path, mode):
         file with.
     """
     ext = path.rpartition('.')[-1]
-    if ext not in {'json', 'pkl', 'txt', 'zip'}:
+    if ext not in {'json', 'pkl', 'zip'}:
         raise InvalidArgumentError(
             'Invalid extension. Make sure your filename ends with '
             '.json, .pkl, or .zip.'
@@ -355,7 +356,8 @@ def save(obj, path, mode_pre='w', verbose=True):
     path: str
         File name to save object to. Should end with .txt, .pkl, .zip, or
         .json depending on desired output format. If .zip is used, object will
-        be zipped and then pickled.
+        be zipped and then pickled. (.sh extension is also allowed and will be
+        treated identically to .txt.)
     mode_pre: str
         Determines whether to write or append text. One of ('w', 'a').
     verbose: bool
@@ -369,7 +371,7 @@ def save(obj, path, mode_pre='w', verbose=True):
     path = Path(path)
     os.makedirs(path.parent, exist_ok=True)
     if verbose: print(f'Writing data to {path}.')
-    if path.suffix == '.txt':
+    if path.suffix[1:] in ('txt', 'sh'):
         with path.open(mode_pre) as f:
             f.write(obj)
     else:
@@ -394,7 +396,7 @@ def load(path, verbose=True):
     object: The Python object that was pickled to the specified file.
     """
     path = Path(path)
-    if path.suffix == '.txt':
+    if path.suffix[1:] in ('txt', 'sh'):
         return path.read_text()
 
     opener, mode, saver = _read_write_args(str(path), 'r')
@@ -725,7 +727,12 @@ class BasicPipeline:
         return x
 
     def __repr__(self):
-        return f'BasicPipeline({", ".join(func_name(f) for f in self.funcs)})'
+        # Try to display each item in the form that was likely passed in: for
+        # functions, this is the name, but for callable classes this is
+        # the str representation of the object, not the class itself.
+        names = ',\n\t'.join(str(f) if hasattr(f, '__call__') else func_name(f)
+                             for f in self.funcs)
+        return f'{type(self).__name__}(\n\t{names}\n)'
 
 
 def pipe(x, *funcs, verbose=False, attr=''):
@@ -897,6 +904,26 @@ def smap(*x):
     return amap('shape', *x)
 
 
+def hasstatic(cls, meth_name):
+    """Check if a class possesses a staticmethod of a given name. Similar to
+    hasattr
+
+    Parameters
+    ----------
+    cls: Type or any
+        A class or an instance (seems to work on both, though more extensive
+        testing may be needed for more complex scenarios).
+    meth_name: str
+        Name of method to check. If the class/instance does not contain any
+        attribute with this name, function returns False.
+
+    Returns
+    -------
+    bool: True if `cls` has a staticmethod with name `meth_name`.
+    """
+    return isinstance(getattr_static(cls, meth_name, None), staticmethod)
+
+
 def parallelize(func, items, total=None, chunksize=1_000, processes=None):
     """Apply a function to a sequence of items in parallel. A progress bar
     is included.
@@ -920,7 +947,7 @@ def parallelize(func, items, total=None, chunksize=1_000, processes=None):
 
     Returns
     -------
-
+    list
     """
     total = total or len(items)
     with Pool(processes) as p:
@@ -980,24 +1007,37 @@ def listlike(x):
     return isinstance(x, Iterable) and not isinstance(x, (str, Mapping))
 
 
-def tolist(x, strict=False):
+def tolist(x, length_like=None, length=None,
+           error_message='x length does not match desired length.'):
     """Helper to let a function accept a single value or a list of values for
-    a certain parameter. See examples.
+    a certain parameter.
+
+    WARNING: if x is a primitive and you specify a length (either via
+    `length_like` or `length`, the resulting list will contain multiple
+    references to the same item). This is mostly intended for use on lists of
+    floats or ints so I don't think it's a problem, but keep this in mind when
+    considering using this on mutable objects.
 
     Parameters
     ----------
     x: Iterable
-        Usually an object that could either be a list/tuple or a primitive,
-        depending on what user passed in.
-    strict: bool
-        If True, returned value will always be a list. If False, we allow
-        tuples/sets/etc. to retain their initial type.
+        Usually either a list/tuple or a primitive.
+    length_like: None or object
+        If provided, we check that x is the same length. If x is a primitive,
+        we'll make it the same length.
+    length: None or int
+        Similar to `length_like` but lets us specify the desired length
+        directly. `length_like` overrides this, though you should only provide
+        one or the other.
+    error_message: str
+        Displayed in the event that a desired length is specified and x is
+        list-like and does not match that length. You can pass in your own
+        error message if you want something more specific to your current use
+        case.
 
     Returns
     -------
-    Iterable: list if strict is True or if x is a primitive. If strict is
-    False and x is already a tuple/set/something similar, its type will be
-    retained.
+    list
 
     Examples
     --------
@@ -1005,16 +1045,26 @@ def tolist(x, strict=False):
         lrs = tolist(lrs)
         ...
 
+    We can now pass in a single learning rate or multiple.
     >>> train(3e-3)
     >>> train([3e-4, 3e-3])
     """
+    if length_like is not None: length = len(length_like)
+
+    # Case 1. List-like x
+    if listlike(x):
+        if length:
+            assert len(x) == length, error_message
+        return list(x)
+
+    # Case 2. Dict-like x
     if isinstance(x, Mapping):
         raise ValueError('x must not be a mapping. It should probably be a '
                          'primitive (str, int, etc.) or a list-like object '
                          '(tuple, list, set).')
-    elif isinstance(x, Iterable) and not isinstance(x, str):
-        return list(x) if strict else x
-    return [x]
+
+    # Case 3. Primitive x
+    return [x] * (length or 1)
 
 
 def xor_none(*args, n=1):
@@ -1234,6 +1284,49 @@ def camel2snake(text):
     return ''.join(res)
 
 
+def to_snake(text):
+    """Experimental feature: tries to convert any common format to snake case.
+    This hasn't been extensively tested but it seems to work with snake case
+    (no change), camel case, upper camel case, words separated by
+    hyphens/dashes/spaces, and combinations of the above. It may occasionally
+    split words that should not be split, though this should be rare if names
+    use actual English words (this might not work so well on fastai-style
+    variable names (very short, e.g. "tfms" for "transforms"), but the intended
+    use case is mostly for fixing column names in pandas.
+
+    Parameters
+    ----------
+    text: str
+
+    Returns
+    -------
+    str: Input text converted to snake case.
+    """
+    return '_'.join(wn.split(text.lower()))
+
+
+def to_camel(text):
+    """Experimental feature: tries to convert any common format to camel case.
+    This hasn't been extensively tested but it seems to work with camel case
+    (no change), snake case, upper camel case, words separated by
+    hyphens/dashes/spaces, and combinations of the above. It may occasionally
+    split words that should not be split, though this should be rare if names
+    use actual English words (this might not work so well on fastai-style
+    variable names (very short, e.g. "tfms" for "transforms"), but the intended
+    use case is mostly for fixing column names in pandas.
+
+    Parameters
+    ----------
+    text: str
+
+    Returns
+    -------
+    str: Input text converted to snake case.
+    """
+    return ''.join(w.title() if i > 0 else w
+                   for i, w in enumerate(wn.split(text.lower())))
+
+
 def kwargs_fallback(self, *args, assign=False, **kwargs):
     """Use inside a method that accepts **kwargs. Sometimes we want to use
     an instance variable for some computation but want to give the user the
@@ -1324,4 +1417,15 @@ def cd_root(root_subdir='notebooks'):
     print('Current directory:', os.getcwd())
 
 
+def ngrams(word, n=3, step=1, drop_last=False):
+    """To get non-overlapping sequences, pass in same value for `step` as `n`.
+    """
+    stop = max(1, step+len(word)-n)
+    ngrams_ = []
+    for i in range(0, stop, step):
+        ngrams_.append(word[i:i+n])
+    if drop_last and len(ngrams_[-1]) < n: ngrams_ = ngrams_[:-1]
+    return ngrams_
+
+
 SENTINEL = object()
